!
! A MODULE 
!                               
! "THE BEER-WARE LICENSE":
! Alberto Ramos wrote this file. As long as you retain this 
! notice you can do whatever you want with this stuff. If we meet some 
! day, and you think this stuff is worth it, you can buy me a beer in 
! return. <alberto.ramos@cern.ch>
!
! 20070110 
!                               
MODULE Time

  USE NumTypes
!  USE Error

  IMPLICIT NONE

  ! The year the Gregorian calendar was implemented 
  ! for more details take a look to
  ! http://en.wikipedia.org/wiki/Gregorian_calendar
  Integer, Parameter :: GREGORIAN_DATE = 1582

  Character (len=3), Dimension  (0:6), Parameter :: NAMES_OF_DAYS = &
       & (/'Sun', 'Mon', 'Tue', 'Wed', 'Thu', 'Fri', 'Sat'/)

  Character (len=3), Dimension (0:11), Parameter :: NAMES_OF_MONTHS = &
       & (/'Jan', 'Feb', 'Mar', 'Apr', 'May', 'Jun', &
       &   'Jul', 'Aug', 'Sep', 'Oct', 'Nov', 'Dec'  /)

  Type tm
     Integer :: hour  ! Hour of the day [0-23]
     Integer :: min   ! Minutes after the hour [0-59]
     Integer :: sec   ! Seconds after the minute [0-59]
     Integer :: msec  ! Miliseconds after the second [0-999]
     Integer :: mday  ! Day of the month [1-31]
     Integer :: mon   ! Month of the year [0-11]
     Integer :: year  ! Year 
     Integer :: wday  ! Day of the week since sunday [0-6]
  End Type tm
  

  Private NAMES_OF_DAYS, NAMES_OF_MONTHS, GREGORIAN_DATE

CONTAINS

! ******************************************************
! *
  Integer Function Day_of_Week(Day, Month, Year)
! *
! ******************************************************
! * Computes the day of the week using the Zeller's
! * congruence. More details at:
! * http://en.wikipedia.org/wiki/Zeller%27s_congruence
! ******************************************************
    Integer, Intent (in) :: Day, Month, Year
    Integer :: Mon, yy, cc, cyear
    
    If (Month == 0) Then
       Mon = 13
       cyear = year - 1
    Else If (Month == 1) Then
       Mon = 14
       cyear = year - 1
    Else 
       Mon = Month + 1
       cyear = year
    End If
    yy = Mod(cyear, 100)
    cc = Int((cyear - yy)/100)

    If (Year < GREGORIAN_DATE) Then
       Day_Of_Week = Day + Int(26*(mon+1)/10) + Int(5*yy/4) - &
            & cc + 5
    Else
       Day_Of_Week = Day + Int(26*(mon+1)/10) + Int(5*yy/4) + &
            & Int(cc/4) - 2*cc 
    End If
    Day_Of_Week = Modulo(Day_Of_Week-1,7)


    Return
  End Function Day_of_Week

! ***************************************
! *
  Type (tm) Function gettime()
! *
! ***************************************
    
    Character (len=8)  :: Datestr
    Character (len=10) :: Timestr
    Character (len=1) :: ch

    CALL Date_and_Time(Datestr, Timestr)

    Read(Datestr,'(1I4,1I2,1I2)')Gettime%year, &
         & Gettime%mon, Gettime%mday
    Read(Timestr,'(1I2,1I2,1I2,1A1,1I3)')Gettime%hour, &
         & Gettime%min, Gettime%sec, ch, Gettime%msec

    Gettime%mon = Gettime%mon - 1
    
    Gettime%wday = Day_of_week(Gettime%mday, Gettime%mon, Gettime%year)

    Return
  End Function gettime

! ***************************************
! *
  Character (len=24) Function asctime(t)
! *
! ***************************************
    
    Type (tm), Intent (in) :: t
  
10  FORMAT((4A,1I2,1A,1I2,1A1,1I2.2,1A1,1I2.2,1A1,1I4))
    
    Write(asctime,10)NAMES_OF_DAYS(t%wday), ' ', &
         & NAMES_OF_MONTHS(t%mon), ' ', t%mday,' ', &
         & t%hour, ':', t%min, ':', t%sec, ' ', t%year
    
    Return
  End Function asctime

  


! ***************************************
! *
  Logical Function isleap(Nyear)
! *
! ***************************************
! * Returns .true. if the year is leap,
! * .false. in any other case
! ***************************************
    
    Integer, Intent (in) :: Nyear

    If (Nyear < GREGORIAN_DATE) Then
       If (Mod(Nyear,4) == 0) Then
          isleap = .True.
       Else
          isleap = .False.
       End If
    Else
       If ( (Mod(Nyear, 400) == 0) .or. &
            & ((Mod(Nyear,4) == 0) .and. (Mod(Nyear,100) /= 0))) Then
          isleap = .True.
       Else
          isleap = .False.
       End If
    End If
          
    Return
  End Function isleap


! ***************************************
! *
  Real (kind=DP) Function Julian_Date(t)
! *
! ***************************************
! ***************************************
    
    Type (tm), Intent (in) :: t
    
    Integer :: a, y, m, JDN
    
    a = Int( (15-t%mon)/12)
    y = t%year + 4800 - a
    m = t%mon + 12*a - 2
    
    ! First compute Julian Day Number
    JDN = t%mday + Int((153*m + 2)/5) + 365*y + Int(y/4) - &
         & Int(y/100) + Int(y/400) - 32045
    
    Julian_Date = Real(JDN,kind=DP) + &
         & Real(t%hour - 12,kind=DP)/24.0_DP + &
         & Real(t%min, kind=DP)/1440.0_DP    + &
         & Real(t%sec, kind=DP)/86400.0_DP   + &
         & Real(t%msec, kind=DP)/86400000.0_DP 
    
    Return
  End Function Julian_Date

! ********************************
! *
    function unix2date(utime) result (t)
! *
! ********************************

      Integer, Intent (in) ::  utime
      type (tm) :: t
      Integer :: idate(6)
      
      !*utime  input  Unix system time, seconds since 1970.0
      !*idate  output Array: 1=year, 2=month, 3=date, 4=hour, 5=minute, 6=secs
      !*-Author  Clive Page, Leicester University, UK.   1995-MAY-2
      Integer :: mjday, nsecs
      Real (kind=4) :: day
      
      !*Note the MJD algorithm only works from years 1901 to 2099.

      mjday   = int(utime/86400 + 40587)
      t%year  = 1858 + int( (mjday + 321.51) / 365.25)
      day     = aint( mod(mjday + 262.25, 365.25) ) + 0.5
      t%mon   = int(mod(day / 30.6 + 2.0, 12.0) )
      t%mday  = 1 + int(mod(day,30.6))
      nsecs   = mod(utime, 86400)
      t%sec   = mod(nsecs, 60)
      nsecs   = nsecs / 60
      t%min   = mod(nsecs, 60)
      t%hour  = nsecs / 60
      t%wday  = Day_of_Week(t%mday, t%mon, t%year)

      Return
    End function unix2date
  

End MODULE Time
