!
! A PROGRAM for 
!                               
! "THE BEER-WARE LICENSE":
! Alberto Ramos wrote this file. As long as you retain this 
! notice you can do whatever you want with this stuff. If we meet some 
! day, and you think this stuff is worth it, you can buy me a beer in 
! return. <alberto.ramos@cern.ch>
!
! 20180606 
!                               
! Reads input file and mean/error

program uwerr

  use ISO_FORTRAN_ENV, Only : error_unit, output_unit, input_unit, iostat_end
  use numtypes
  use constants
  use aderrors
  use nonnumeric
  
  implicit none

  real (kind=DP), allocatable :: data(:)
  real (kind=DP)              :: stau, texp, dfoo, dsig
  type (uwreal)               :: obs
  integer                     :: nsz, ifin, ifout, ist, nskip, i, nr
  integer, allocatable        :: ivr(:)
  character (len=1000)        :: fin, fout, cfoo
  
  if (getopt('-h')) call help()
  if (.not.getopt('-i', fin)) call help()
  open(newunit=ifin, file=trim(fin))

  if (.not.getopt('-skip',nskip)) nskip = 0
  if (.not.getopt('-texp', texp)) texp  = 0.0_DP
  if (.not.getopt('-stau', stau)) stau  = 4.0_DP
  if (.not.getopt('-dsig', dsig)) dsig  = 1.5_DP
  
  nsz   = 0
  do
     read(ifin,*,iostat=ist)dfoo
     if (ist.eq.iostat_end) exit
     nsz = nsz + 1
  end do
  allocate(data(nsz-nskip))
  
  rewind(ifin)
  do i = 1, nskip
     read(ifin,*)dfoo
  end do
  do i = 1, nsz-nskip
     read(ifin,*)data(i)
  end do
  close(ifin)

  if (getopt('-rep',cfoo)) then
     nr = 1
     do i = 1, len_trim(cfoo)
        if (cfoo(i:i)==',') nr=nr+1
     end do

     allocate(ivr(nr))
     read(cfoo,*)ivr
  else
     allocate(ivr(1))
     ivr = nsz-nskip
  end if

  
  obs = data
  call obs%set_stau(stau)
  call obs%set_texp(texp)
  call obs%set_dsig(dsig)
  call obs%set_replica(ivr)
  call obs%uwerr()

  if (getopt('-o',fout)) then
     open(newunit=ifout,file=trim(fout))
     call obs%print_hist(ifout)
     close(ifout)
  end if
  
  write(output_unit,'(3X,1A,1I8, 1A)')'# Read in ', nsz, ' data'
  write(output_unit,'(3X,"# stau: ",1F5.2,5X,"tau_exp:",1F8.2)')stau, texp
  write(output_unit,'(3X,1A,1F0.12," +/- ",1ES10.4)')'Mean:    ', &
       obs%value(), obs%error()
  write(output_unit,'(3X,1A,1F0.12," +/- ",1ES10.4, 1A, 1I4,1A)')&
       'tau_int: ', obs%taui(1), obs%dtaui(1), ' (window: ', obs%window(1), ')'

  stop
contains

  subroutine help()
    
    character (len=1000) :: prg

    call get_command_argument(0,prg)
    write(error_unit,*)'usage: '
    write(error_unit,*)trim(prg), ' -i <data file> [options]'
    write(error_unit,*)
    write(error_unit,*)'  Options: '
    write(error_unit,*)'    -skip <n>:            Skip first n lines from input file '
    write(error_unit,*)'    -texp <tau_exp>:      Set exponential autocorrelation time for Tails '
    write(error_unit,*)'    -dsig <tail_attach>:  Set where to attach the tail (standard deviations)'
    write(error_unit,*)'    -stau <S_tau>:        Set S_tau for automatic windowing procedure    '
    write(error_unit,*)'    -rep  <r1,r2,...,rn>: Set replica vector                             '
    write(error_unit,*)'    -o    <fname>:        Set file name for detailed output              '
    write(error_unit,*)
    write(error_unit,*)'  Examples: '
    write(error_unit,*)
    write(error_unit,*)'  ', trim(prg), ' -i data.dat -texp 100.0 -dsig 2.0' 
    write(error_unit,'(100(7X,1A))')'Analyze data from file data.dat with an exponential '
    write(error_unit,'(100(7X,1A))')'autocorrelation time of 100 (for tails). Attach the '
    write(error_unit,'(100(7X,1A))')'tail when the signal is two standard deviations away '
    write(error_unit,'(100(7X,1A))')'from zero'
    write(error_unit,*)
    write(error_unit,*)'  ', trim(prg), ' -i data.dat -stau 4.0 -o data.out' 
    write(error_unit,'(100(7X,1A))')'Analyze data from file data.dat with Stau=4 for automatic '
    write(error_unit,'(100(7X,1A))')'window procedure. Write detailed output to file data.out'
    write(error_unit,*)
    write(error_unit,*)'  ', trim(prg), ' -i data.dat -rep "1000,2500,500" -skip 100' 
    write(error_unit,'(100(7X,1A))')'Analyze data from file data.dat in three repicas of sizes '
    write(error_unit,'(100(7X,1A))')'1000,2500,500 with default parameters skipping the first '
    write(error_unit,'(100(7X,1A))')'100 lines of the file. (The file must contain 4100 lines)'
   
    
    error stop
  end subroutine help
  
end program uwerr
